#! /bin/bash
# The above is a lie to keep Emacs flycheck-mode (which uses ShellCheck)
# happy and useful, because flycheck-mode thinks that /bin/sh is POSIX not
# BASH.  This script is actually processed by /bin/sh, which is currently
# BASH 3.2 on all macOS platforms, and not the much more modern /bin/zsh,
# because the scripts which use this code are used on non-macOS platforms.
# Thank you, Apple, for not updating /bin/sh (same as /bin/bash) since 2006.
#
# This script is used by ACL's update.sh and app startup scripts
# (ACL, Gruff).  
#
# This script is source'd not executed.  The job of this script is to:
#  1. side effect PATH, if needed, to include the directory
#     containing the `openssl` command,
#  2. optionally export ACL_OPENSSL_VERSION, and
#  3. set the BASH variable "___env" to an array of variable/value pairs
#     or null.
# The caller of this script can then execute ACL or the application with
#   "${___env[@]}" lisp ...
# so that `(require :ssl)` will work.
#
# The varied SSL conditions we check for, in this order:
#  * OpenSSL included with the application
#    - the file ACL_OPENSSL_VERSION will contain the value for the
#      environment variable of the same name, which is then exported
#  * Homebrew OpenSSL
#    - broken "brew --prefix openssl" (points to non-existent directory)
#    - working installing
#  * Macports OpenSSL
#
# If none of the conditions above are met, then we exit with an error
# message, because the ACL SSL module does NOT work with Apple's SSL.
# That is LibreSSL, which probably fine and is compatible with OpenSSL 1.0,
# but Apple does not provide include files so we can compiled against it.
#
# This script returns TRUE if all goes well, otherwise it will
# print an error message to stderr and return FALSE.   It is incumbent
# on the caller to do something like this:
#    if ! source thisscript.sh; then
#       ...error condition...

SCRIPTDIR=$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )

declare -a ___env
___env=()

###############################################################################
# OpenSSL included

if [ -f "$SCRIPTDIR/ACL_OPENSSL_VERSION" ]; then
    ACL_OPENSSL_VERSION="$(cat "$SCRIPTDIR/ACL_OPENSSL_VERSION")"
    export ACL_OPENSSL_VERSION
    # make sure we lookin $SCRIPTDIR for OpenSSL first, because
    # there may be an installation of Macports/Homebrew OpenSSL,
    # but it could be a different version.
    ___env+=("DYLD_LIBRARY_PATH=$SCRIPTDIR")
    return 0
fi

###############################################################################
# Homebrew

if __homebrew=$(type -p brew); then
    :
elif [ -x /opt/homebrew/bin/brew ]; then
    __homebrew=/opt/homebrew/bin/brew
elif [ -x /usr/local/bin/brew ]; then
    __homebrew=/usr/local/bin/brew
else
    __homebrew=
fi

if [ "$__homebrew" ]; then
    __ssldir=$($__homebrew --prefix openssl)
    if [ ! -d "$__ssldir" ]; then
        __ssldir=$($__homebrew --prefix openssl@1.1)
        if [ ! -d "$__ssldir" ]; then
            echo "Error: Homebrew openssl and openssl@1.1 do not exist." 1>&2
            return 1
        fi
    fi

    ___env+=("DYLD_LIBRARY_PATH=${__ssldir}/lib")
    PATH="${__ssldir}/bin:$PATH"
    
    return 0
fi

###############################################################################
# Macports

if __macports=$(type -p port); then
    __macportsdir=$(echo "$__macports" | sed 's,/bin/port,,')
elif [ -x /opt/local/bin/port ]; then
    __macports=/opt/local/bin/port
    __macportsdir=/opt/local
fi

if [ "$__macports" ]; then
    ___env+=("DYLD_LIBRARY_PATH=${__macportsdir}/lib")
    PATH="${__macportsdir}/bin:$PATH"

    return 0
fi

###############################################################################
echo "Error: could not find OpenSSL" 1>&2
return 1
